
/********************************************************************************************************
 * mm use Header and footer for each block allocated/non allocated
 * footer helps to insert in free list in 0(1) time
 * also two-sided coalescing can be done in 0(1) time
 * freelist uses double-linked list
 * if either or both adjacent block are free it willbe merged with them
 * else block will be attached to head of list
 * free call will take o(1) time for coalescing andd insertion
 * MM_malloc first search for a block in free list if avilable using best first it will return
 * if not available new block will be requested from mem_sbrk
 
 *Metadata size in allocated block  is 8(footer)+8(header)=16 byte 
 * min block size will be !6 +8(min size after align)=24 byte
 *metadata size for free block will be 16(header)+8(footer)=24 byte
 *since min block size is guaranted to be 24 hence metadata for free block can be stored easily 
 *in giveen space
 
 ********************************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>

#include "mm.h"
#include "memlib.h"

team_t team = {
    /* Team name */
    "team",
    /* First member's full name */
    "Paras Garg",
    /* First member's email address */
    "parasgarg@cse.iitb.ac.in",
    /* Second member's full name (leave blank if none) */
    "member 2",
    /* Second member's email address (leave blank if none) */
    "member_2@cse.iitb.ac.in"
};

/* single word (4) or double word (8) alignment */
#define ALIGNMENT 8

/* rounds up to the nearest multiple of ALIGNMENT */
#define ALIGN(size) (((size) + (ALIGNMENT-1)) & ~0x7)


#define SIZE_T_SIZE (ALIGN(sizeof(size_t)))


#define ALLOC_UNIT sysconf(_SC_PAGESIZE)
//size of struct on 32bit system is 8 byte
typedef struct blk_header_t	{
	size_t size;
	int free;
	}blk_header;
//size of struct on 32bit system is 16 byte		
typedef struct free_blk_header_t	{
	size_t size;
	int free;
	struct free_blk_header_t *prev;
	struct free_blk_header_t *next;
	}free_blk_header;
//size of struct on 32 bit system is 8 byte	
typedef struct blk_footer_t	{
	int free;
	size_t size;
	}blk_footer;
   	
#define METADATA_SIZE (sizeof(blk_footer)+sizeof(blk_header))

free_blk_header *head;
char* MAX_ADDRESS;
char* MIN_ADDRESS;
void *init_mem_sbrk_break = NULL;

/* 
 * mm_init - initialize the malloc package.
 */
int mm_init(void)
{
	head=NULL;
	MIN_ADDRESS=mem_sbrk(0);
	//printf("ini%p",(void*)MIN_ADDRESS);
	//printf("starting address %p",(void*)MIN_ADDRESS);
	MAX_ADDRESS=MIN_ADDRESS;
	/*
	 * IT resest the head of freelist to null
	 * memlib reset the heap section used resetting head do the work	

	 */
	
    return 0;		//Returns 0 on successfull initialization.
}

//---------------------------------------------------------------------------------------------------------------
/* 
 * mm_malloc - Allocate a block by incrementing the brk pointer.
 *     Always allocate a block whose size is a multiple of the alignment.
 */
void *mm_malloc(size_t size)
{	
	/* 
	 * This function should keep track of the allocated memory blocks.
	 * The block allocation should minimize the number of holes (chucks of unusable memory) in the heap memory.
	 * The previously freed memory blocks should be reused.
	 * If no appropriate free block is available then the increase the heap  size using 'mem_sbrk(size)'.
	 * Try to keep the heap size as small as possible.
	 */
	//printf("\n allocating %d",size);
	if(size <= 0){		// Invalid request size
		return NULL;
	}
	size=ALIGN(size);
	int requiredSize=size+METADATA_SIZE;
	
	
	free_blk_header *temp=head;
	free_blk_header *bestfit=NULL;
	while(temp!=NULL){
		//	printf("\n freelist search block address %p size %d",(void*)temp,temp->size);
			fflush(stdout);
		if(temp->size >= requiredSize){

			if(bestfit==NULL){
				bestfit=temp;
				}
			else if(bestfit->size>temp->size){
				bestfit=temp;
				}
		}
		temp=temp->next;	  
	}
	if(bestfit!=NULL){
	//because block size may be greater and hence 
	requiredSize=bestfit->size;
	//printf("\n block mil gaya");
	free_blk_header *prev=bestfit->prev;
	temp=bestfit->next;
		if(prev==NULL){
			if(temp==NULL){
				head=NULL;
			}else{
				head=temp;
				temp->prev=NULL;
			}
		}else{
			if(temp==NULL){
				prev->next=NULL;
			}else{
			prev->next=temp;
			temp->prev=prev;
			}
		}
	}
	//size = ((size+7)/8)*8;		//size alligned to 8 bytes

	//printf("\nrequiredSize %d ",size);
	void *ptr;
	if(bestfit==NULL){
		ptr=mem_sbrk(requiredSize);
		MAX_ADDRESS=(char*)ptr+requiredSize;
	}
	else{
	ptr=(void*)bestfit;
	}
	
	//printf(" %d %d %p",sizeof(blk_footer),sizeof(blk_header),ptr+size+2);
	
	//set header metadata
	blk_header* header=(blk_header*)ptr;
	//printf("\nheader %p ",ptr);
	header->size=requiredSize;
	header->free=0;
	
	//printf("\n size %d allocated %d",header->size,header->free);
	//set footer metadata
	blk_footer* footer=(blk_footer*)((char*)ptr+requiredSize-sizeof(blk_footer));
	//printf("\nheader %p size %d free %dfooter size %d free %d",(void*)header,header->size,header->free,footer->size,footer->free);
	footer->size=requiredSize;
	footer->free=0;
	
	//returing starting address for user data
	return (void*)((char*)ptr+sizeof(blk_header));	
	//mem_sbrk() is wrapper function for the sbrk() system call. 
	//Please use mem_sbrk() instead of sbrk() otherwise the evaluation results 
	//may give wrong results
}
void bothfree(blk_header* left,blk_header* curr,blk_header* righthead){

	free_blk_header *right=(free_blk_header*)righthead;
	//Arrange pointer for right side
	if(right->prev!=NULL && right->next!=NULL){
		right->prev->next=right->next;
		right->next->prev=right->prev;
		
	}else if(right->prev==NULL){
		head=right->next;
		right->next->prev=NULL;
	}else if(right->next==NULL){
		right->prev->next=NULL;
	}
		
	//now increase left size to leftsize+size+rightsize
	left->size=left->size+curr->size+right->size;
	blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
	footer->free=1;
	footer->size=left->size;

}
void bothNotFree(blk_header *header){
	int size=header->size;
	free_blk_header *free=(free_blk_header*)((void*)header);
	free->size=size;
	free->free=1;
	free->prev=NULL;
	free->next=NULL;
	//Attach block to start of list
	if(head==NULL){
		head=free;
	}
	else{
		free->next=head;
		head->prev=free;
		head=free;
	}
}
void leftFree(blk_header* left,blk_header* head){
	//printf("\nleeft free");
	//simply increase block sizze no need to change the pointers
	int size=head->size;
	left->size=left->size+size;
	blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
	footer->free=1;
	footer->size=left->size;
}

void rightFree(blk_header* right,blk_header* currhead){
	//printf("right free");
	int size=currhead->size;
	free_blk_header *prev=((free_blk_header*)right)->prev;
	free_blk_header *next=((free_blk_header*)right)->next;
	int rightsize=right->size;
	//set footer
	blk_footer* footer=(blk_footer*)((char*)right+right->size-sizeof(blk_footer));
	footer->size=rightsize+size;
	
	//set header
	free_blk_header *current=(free_blk_header*)currhead;
	current->free=1;
	current->size=rightsize+size;
	
	current->next=next;
	if(next!=NULL)
		next->prev=current;
	current->prev=prev;
	if(prev!=NULL){
		prev->next=current;
	}else{
	head=current;
	}
}
void mm_free(void *ptr)
{
	 if(ptr==NULL)
	 	return;
	 	
	 blk_header *header=(blk_header*)ptr-1;
	 
	 int left_free=0;
	 int right_free=0;
	 
	 blk_header *rightHead=(blk_header*)((char*)header+header->size);
	 if((char*)rightHead!=MAX_ADDRESS){
	 //printf("right head exists");
	 	if(rightHead->free)
	 		right_free=1;
	 }
	 
	 blk_footer *prev_footer=(blk_footer*)header-1;
	 blk_header *leftHead=(blk_header*)((char*)header-prev_footer->size);
	// printf("prev head %p prev footer%p free %d next head%p free%d",(void*)leftHead,(void*)prev_footer,prev_footer->free,(void*)rightHead,rightHead->free);
	 if((char*)header!=MIN_ADDRESS){
	 //printf("left head exists");
	 	if(leftHead->free)
	 		left_free=1;
	 }
	// printf("\n right starting address  %p size %d free %d",(void*)rightHead,rightHead->size,rightHead->free);
	 //printf("\n left starting address  %p size %d free %d",(void*)leftHead,leftHead->size,leftHead->free);
	/*
	 * finding both adjacents blocks
	 * and check whether both are free
	*/
	if(left_free==1 && right_free==1){
		bothfree(leftHead,header,rightHead);
	}
	if(left_free==0 && right_free==1){
		rightFree(rightHead,header);
		}
	if(left_free==1 && right_free==0){
		leftFree(leftHead,header);
		}
		
	if(left_free==0 && right_free==0){
		bothNotFree(header);
	 }
}
void printmem(){
	blk_header *header;
	char* temp=MIN_ADDRESS;
	printf("\n printing mem %p",(void*)MIN_ADDRESS);
	fflush(stdout);
	int i=1;
	while(temp<MAX_ADDRESS){
		header=(blk_header*)((void*)temp);
		temp=temp+header->size;
		printf("\n Block %d Size %d free %d %p %p",i++,header->size,header->free,(void*)temp,(void*)MAX_ADDRESS); 
	
	}
}

/*
 * mm_realloc - Implemented simply in terms of mm_malloc and mm_free
 */
void *mm_realloc(void *ptr, size_t size)
{	
	if(ptr == NULL){			//memory was not previously allocated
		return mm_malloc(size);
	}
	
	if(size == 0){				//new size is zero
		mm_free(ptr);
		return NULL;
	}

	/*
	 * This function should also copy the content of the previous memory block into the new block.
	 * You can use 'memcpy()' for this purpose.
	 * 
	 * The data structures corresponding to free memory blocks and allocated memory 
	 * blocks should also be updated.
	*/
	
	void* memory=mm_malloc(size);
	memcpy(memory,ptr,((blk_header*)ptr-1)->size);
	mm_free(ptr);
	return memory;
	
}


/*
int main(){

mem_init();
mm_init();

void *temp0=mm_malloc(4);
//printmem();

void *temp=mm_malloc(4);
//printmem();
 void* temp2=mm_malloc(4);
 //printmem();
mm_free(temp2);
//printmem();

mm_malloc(7);
//printmem();
mm_free(temp2);
mm_free(temp);
//printmem();
mm_realloc(temp0,14);
//printmem();
}
*/











